from control import TransferFunction
import numpy as np

def extract_poles_and_zeros(sys):
    """
    Extract poles and zeros from a TransferFunction object.
    """
    poles = np.roots(sys.den[0][0])
    zeros = np.roots(sys.num[0][0])
    return np.sort_complex(poles), np.sort_complex(zeros)

def factor_poles(poles):
    """
    Factor poles into sub-TF models.
    """
    F = []
    F_gain = 1

    for pole in poles:
        if np.iscomplex(pole):
            F.append(TransferFunction([1], [1, -2 * np.real(pole), np.abs(pole)**2 + np.imag(pole)**2]))
            F_gain /= np.abs(pole)**2
        else:
            F.append(TransferFunction([1], [1, -pole]))
            F_gain /= np.abs(pole)

    return F, F_gain

def factor_zeros(zeros):
    """
    Factor zeros into sub-TF models.
    """
    F = []
    F_gain = 1

    for zero in zeros:
        if np.iscomplex(zero):
            F.append(TransferFunction([1, -2 * np.real(zero), np.abs(zero)**2 + np.imag(zero)**2], [1]))
            F_gain *= np.abs(zero)**2
        else:
            F.append(TransferFunction([1], [1, -zero]))
            F_gain *= np.abs(zero)

    return F, F_gain

def tf_factor(sys):
    """
    Factor a transfer function into constant, real pole/zero, and 
    conjugate pole/zero pair sub-TF models.
    """
    if not isinstance(sys, TransferFunction):
        raise ValueError("Input 'sys' must be a TransferFunction object.")

    poles, zeros = extract_poles_and_zeros(sys)
    F_poles, F_poles_gain = factor_poles(poles)
    F_zeros, F_zeros_gain = factor_zeros(zeros)

    F = F_poles + F_zeros + [TransferFunction([F_poles_gain * F_zeros_gain], [1])]

    # Check by concatenation
    tf_composite = TransferFunction([1], [1])
    for sub_tf in F:
        tf_composite *= sub_tf

    print(tf_composite)
    
    num_gain = sys.num[0][0][0]
    den_gain = sys.den[0][0][0]

    if np.allclose(num_gain / den_gain, F_poles_gain * F_zeros_gain):
        return F
    else:
        raise ValueError("Composite check failed!")

# Example usage:
if __name__ == "__main__":
    sys = TransferFunction([-0.64, -0.4101, 0.00783], [1, 1.489, 0.7681, 0.09455, 0.0424, 0.7])
    tf_factors = tf_factor(sys)
    print(tf_factors)



import control
from scipy.signal import tf2zpk
import numpy as np
import numpy.polynomial.polynomial as poly

class TransferFunction(control.TransferFunction):
    """Subclass of control.TransferFunction with extra methods"""
    def pop_conjugate(self, root, roots):
        """Pop the conjugate of a root in roots list
        
        Finds the closest root in roots to the complex conjugate of root.
        Uses the "closest" metric of the Euclidean distance.
        """
        rootsa = np.array(roots)
        diff = root.conjugate() - rootsa
        dist = np.abs(diff)
        imini = dist.argmin()
        mini = roots.pop(imini)
        return mini
    
    def poly_factors_canonical(self, p):
        """Returns polynomial factors in canonical form"""
        p = poly.Polynomial(np.flip(p))  # Polynomials have increasing powers
        n = p.degree()
        roots = list(p.roots())
        K = p.coef[-1]
        factors = []
        while roots:
            root = roots.pop()
            if np.iscomplex(root):
                root_conj = self.pop_conjugate(root, roots)
                factor = poly.polyfromroots([root, root_conj])  # $s^2 + 2\zeta\omega_n s + \omega_n^2$
                factor = np.flip(np.real(factor))  # Should all be real
                k = 1/factor[-1]  # Factor gain $1/w_n^2$
                K = K/k  # Overall gain absorbing factor gain
            else:
                factor = poly.polyfromroots([root])
                factor = np.flip(np.real(factor))  # $s + 1/\tau$
                print(factor)
                K = K*factor[-1]  # Overall gain absorbing $\tau$
                factor = factor/factor[-1]  # $\tau s + 1$
                k = 1  # Factor gain
            factors.append((k, factor))
        factors.insert(0, (K, np.array([1])))  ## Prepend the gain as the first factor
        return factors
        
    def factor_canonical(self, check=False):
        """Returns a list of transfer functions in canonical form, the product of which equals self."""
        if self.ninputs > 1 or self.noutputs > 1:
            raise NotImplementedError(
                "TransferFunction.factor_canonical is currently only implemented "
                "for SISO systems.")
        zpf = self.poly_factors_canonical(np.array(self.num).flatten())  # Zero factors
        ppf = self.poly_factors_canonical(np.array(self.den).flatten())  # Pole factors
        gain = zpf.pop(0)[0]/ppf.pop(0)[0]
        tf_factors = [control.TransferFunction([gain],[1])]
        for zf in zpf:
            tf_factors.append(
                control.TransferFunction(zf[1], [1/zf[0]])
            )
        for pf in ppf:
            tf_factors.append(
                control.TransferFunction([1/pf[0]], pf[1])
            )
        if check:
            tf_from_factors = 1
            for tf in tf_factors:
                tf_from_factors *= tf
            tf_from_factors = tf_from_factors.minreal()
            assert(
                np.allclose(
                    np.array(tf_from_factors.num).flatten(), 
                    np.array(self.num).flatten().astype(np.float64)
                )
            )
            assert(
                np.allclose(
                    np.array(tf_from_factors.den).flatten(),
                    np.array(self.den).flatten().astype(np.float64)
                )
            )
        return tf_factors
        
        
        


import sympy as sp

class TransferFunctionSymbolic:
    """Represents a continuous LTI transfer function model in symbolic form"""

    def __init__(self, H, s=None):
        self.H = H
        if s is not None:
            self.s = s
        else:
            self.s = list(H.free_symbols)[0]
    
    def factor_p(self, p, poles=True):
        p = sp.Poly(p, self.s).factor_list()  # Factored
        if poles:
            K = p[0]  # Overall gain
        else:
            K = 1/p[0]  # Overall gain
        factors = []
        for pi in p[1]:  # Each factor
            m = pi[1]  # Multiplicity
            d = sp.Poly(pi[0], self.s).degree()
            cs_k = sp.Poly(pi[0], self.s).all_coeffs()
            if d == 2:
                k = cs_k[0]  # factor gain
                cs = list(map(lambda c: c/k, cs_k))
                wn2 = cs[2]
                quad = self.s**2 + cs[1]*self.s + cs[2]
                if poles:
                    factor = wn2/quad
                    k = k/wn2
                else:
                    factor = quad/wn2
                    k = wn2/k
            elif d == 1:
                k = cs_k[1]
                tau = cs_k[0]/k
                lin = tau*self.s + 1
                if poles:
                    factor = lin
                else:
                    factor = 1/lin
            elif d == 0:
                factor = 1
                if poles:
                    k = cs_k[0]
                else:
                    k = 1/cs_k[0]
            else:
                raise(RuntimeError(f"Polynomial should not be degree {d}"))
            for _ in range(0, m):
                K = k*K
                factors.append(factor)
        return K, factors

    def factor(self):
        """Returns an overall gain and a list of standard-form terms"""
        num, den = sp.fraction(H.cancel())
        Kz, factors_z = self.factor_p(num, poles=False)
        Kp, factors_p = self.factor_p(den, poles=True)
        K = Kz * Kp
        factors = factors_p + factors_z
        return K, factors

    def to_control(self, params: dict = {}):
        """Returns an equivalent Control Systems package control.StateSpace object"""
        A, B, C, D = self.to_numpy(params=params)
        return control.ss(A, B, C, D)


def tfs(H, s=None):
    """Create a TransferFunctionSymbolic object"""
    return TransferFunctionSymbolic(H, s=s)

# Example usage:
if __name__ == "__main__":
    s = sp.symbols("s")
    H = (2430.0*s + 810.0)/(30.0*s**3 + 271.0*s**2 + 2439.0*s + 81.0)
    sys = tfs(H)
    K, factors = sys.factor()
    print(f"Overall gain: {K}")
    print(f"Factors: {factors}")